
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "gtd-enum-types.h"
/* enumerations from "/build/gnome-todo/src/build/../gnome-todo/src/animation/gtd-animation-enums.h" */
#include "/build/gnome-todo/src/build/../gnome-todo/src/animation/gtd-animation-enums.h"
GType
gtd_ease_mode_get_type (void)
{
	static GType the_type = 0;

	if (the_type == 0)
	{
		static const GEnumValue values[] = {
			{ GTD_CUSTOM_MODE,
			  "GTD_CUSTOM_MODE",
			  "custom-mode" },
			{ GTD_EASE_LINEAR,
			  "GTD_EASE_LINEAR",
			  "ease-linear" },
			{ GTD_EASE_IN_QUAD,
			  "GTD_EASE_IN_QUAD",
			  "ease-in-quad" },
			{ GTD_EASE_OUT_QUAD,
			  "GTD_EASE_OUT_QUAD",
			  "ease-out-quad" },
			{ GTD_EASE_IN_OUT_QUAD,
			  "GTD_EASE_IN_OUT_QUAD",
			  "ease-in-out-quad" },
			{ GTD_EASE_IN_CUBIC,
			  "GTD_EASE_IN_CUBIC",
			  "ease-in-cubic" },
			{ GTD_EASE_OUT_CUBIC,
			  "GTD_EASE_OUT_CUBIC",
			  "ease-out-cubic" },
			{ GTD_EASE_IN_OUT_CUBIC,
			  "GTD_EASE_IN_OUT_CUBIC",
			  "ease-in-out-cubic" },
			{ GTD_EASE_IN_QUART,
			  "GTD_EASE_IN_QUART",
			  "ease-in-quart" },
			{ GTD_EASE_OUT_QUART,
			  "GTD_EASE_OUT_QUART",
			  "ease-out-quart" },
			{ GTD_EASE_IN_OUT_QUART,
			  "GTD_EASE_IN_OUT_QUART",
			  "ease-in-out-quart" },
			{ GTD_EASE_IN_QUINT,
			  "GTD_EASE_IN_QUINT",
			  "ease-in-quint" },
			{ GTD_EASE_OUT_QUINT,
			  "GTD_EASE_OUT_QUINT",
			  "ease-out-quint" },
			{ GTD_EASE_IN_OUT_QUINT,
			  "GTD_EASE_IN_OUT_QUINT",
			  "ease-in-out-quint" },
			{ GTD_EASE_IN_SINE,
			  "GTD_EASE_IN_SINE",
			  "ease-in-sine" },
			{ GTD_EASE_OUT_SINE,
			  "GTD_EASE_OUT_SINE",
			  "ease-out-sine" },
			{ GTD_EASE_IN_OUT_SINE,
			  "GTD_EASE_IN_OUT_SINE",
			  "ease-in-out-sine" },
			{ GTD_EASE_IN_EXPO,
			  "GTD_EASE_IN_EXPO",
			  "ease-in-expo" },
			{ GTD_EASE_OUT_EXPO,
			  "GTD_EASE_OUT_EXPO",
			  "ease-out-expo" },
			{ GTD_EASE_IN_OUT_EXPO,
			  "GTD_EASE_IN_OUT_EXPO",
			  "ease-in-out-expo" },
			{ GTD_EASE_IN_CIRC,
			  "GTD_EASE_IN_CIRC",
			  "ease-in-circ" },
			{ GTD_EASE_OUT_CIRC,
			  "GTD_EASE_OUT_CIRC",
			  "ease-out-circ" },
			{ GTD_EASE_IN_OUT_CIRC,
			  "GTD_EASE_IN_OUT_CIRC",
			  "ease-in-out-circ" },
			{ GTD_EASE_IN_ELASTIC,
			  "GTD_EASE_IN_ELASTIC",
			  "ease-in-elastic" },
			{ GTD_EASE_OUT_ELASTIC,
			  "GTD_EASE_OUT_ELASTIC",
			  "ease-out-elastic" },
			{ GTD_EASE_IN_OUT_ELASTIC,
			  "GTD_EASE_IN_OUT_ELASTIC",
			  "ease-in-out-elastic" },
			{ GTD_EASE_IN_BACK,
			  "GTD_EASE_IN_BACK",
			  "ease-in-back" },
			{ GTD_EASE_OUT_BACK,
			  "GTD_EASE_OUT_BACK",
			  "ease-out-back" },
			{ GTD_EASE_IN_OUT_BACK,
			  "GTD_EASE_IN_OUT_BACK",
			  "ease-in-out-back" },
			{ GTD_EASE_IN_BOUNCE,
			  "GTD_EASE_IN_BOUNCE",
			  "ease-in-bounce" },
			{ GTD_EASE_OUT_BOUNCE,
			  "GTD_EASE_OUT_BOUNCE",
			  "ease-out-bounce" },
			{ GTD_EASE_IN_OUT_BOUNCE,
			  "GTD_EASE_IN_OUT_BOUNCE",
			  "ease-in-out-bounce" },
			{ GTD_EASE_LAST,
			  "GTD_EASE_LAST",
			  "ease-last" },
			{ 0, NULL, NULL }
		};
		the_type = g_enum_register_static (
				g_intern_static_string ("GtdEaseMode"),
				values);
	}
	return the_type;
}
GType
gtd_timeline_direction_get_type (void)
{
	static GType the_type = 0;

	if (the_type == 0)
	{
		static const GEnumValue values[] = {
			{ GTD_TIMELINE_FORWARD,
			  "GTD_TIMELINE_FORWARD",
			  "forward" },
			{ GTD_TIMELINE_BACKWARD,
			  "GTD_TIMELINE_BACKWARD",
			  "backward" },
			{ 0, NULL, NULL }
		};
		the_type = g_enum_register_static (
				g_intern_static_string ("GtdTimelineDirection"),
				values);
	}
	return the_type;
}

/* Generated data ends here */

